/*
 *  Video stacking program with change detection.
 *
 *  Panu Lahtinen, pnuu+finsprite@iki.fi
 */

#include "sky-cam.h"

static void init_peak_hold(void){
  peak_hold->Y = calloc(current_frame.width *
			current_frame.height,
			sizeof(unsigned char));
  if(current_frame.channels == 3){
    peak_hold->Cr = calloc(current_frame.width *
			   current_frame.height / 2,
			   sizeof(unsigned char));
    peak_hold->Cb = calloc(current_frame.width *
			   current_frame.height / 2,
			   sizeof(unsigned char));
  }
  peak_hold->channels = current_frame.channels;
  peak_hold->width = current_frame.width;
  peak_hold->height = current_frame.height;
  peak_hold->timestamp = current_frame.timestamp;
  peak_hold->usec = current_frame.usec;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

static void init_min_hold(void){
  int i = 0;
  min_hold->Y = malloc(current_frame.width*
		       current_frame.height*
		       sizeof(*(min_hold->Y)));
  for(i=0;i<current_frame.width*current_frame.height;i++){
    min_hold->Y[i] = 255;
  }
  if(current_frame.channels == 3){
    min_hold->Cr = malloc(current_frame.width*
			  current_frame.height*
			  sizeof(unsigned char)/2);
    min_hold->Cb = malloc(current_frame.width*
			  current_frame.height*
			  sizeof(unsigned char)/2);
  }
  min_hold->channels = current_frame.channels;
  min_hold->width = current_frame.width;
  min_hold->height = current_frame.height;
  min_hold->timestamp = current_frame.timestamp;
  min_hold->usec = current_frame.usec;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

static void init_latest_img(void){
  latest_img->Y = calloc(current_frame.width*
			 current_frame.height,
			 sizeof(unsigned char));
  if(current_frame.channels == 3){
    latest_img->Cr = calloc(current_frame.width*
			    current_frame.height / 2,
			    sizeof(unsigned char));
    latest_img->Cb = calloc(current_frame.width*
			    current_frame.height / 2,
			    sizeof(unsigned char));
  }
  latest_img->channels = current_frame.channels;
  latest_img->width = current_frame.width;
  latest_img->height = current_frame.height;
  latest_img->timestamp = current_frame.timestamp;
  latest_img->usec = current_frame.usec;
  latest_img_timestamp = current_frame.timestamp;
  latest_img_usec = current_frame.usec;
}

static void init_ave_stack(void){
  ave_stack->Y = calloc(current_frame.width *
			current_frame.height,
			sizeof(unsigned long));
  if(current_frame.channels == 3){
    ave_stack->Cr = calloc(current_frame.width *
			   current_frame.height / 2,
			   sizeof(unsigned long));
    ave_stack->Cb = calloc(current_frame.width *
			   current_frame.height / 2,
			   sizeof(unsigned long));
  }
  ave_stack->channels = current_frame.channels;
  ave_stack->width = current_frame.width;
  ave_stack->height = current_frame.height;
  ave_stack->timestamp = current_frame.timestamp;
  ave_stack->usec = current_frame.usec;
  ave_stack->max_val = 0;
  ave_stack->min_val = ULONG_MAX;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

static void init_md_arrays(void){
  int i, j, n;

  /* Read motion detection mask */
  if(md_mask_fname != NULL){
    md_mask = read_mask(md_mask_fname, &md_mask_width, &md_mask_height);
  }

  /* Check motion detection mask dimensions. If they are not same as
     the video dimension, discard the mask. */
  if(md_mask_width != current_frame.width || 
     md_mask_height != current_frame.height){
    if(md_mask != NULL){
      free(md_mask);
      md_mask = NULL;
    }
  }
  
  /* If there is no mask or it was wrong size, generate dummy mask */
  if(md_mask == NULL){
    md_mask = (unsigned char*)malloc(current_frame.width*current_frame.height*sizeof(unsigned char));
    for(i=0;i<current_frame.width*current_frame.height;i++){ md_mask[i] = 1; }
    md_mask_width = current_frame.width;
    md_mask_height = current_frame.height;
  }

  /* Allocate all motion detection frames in one go */
  if(md_frames == NULL){
    md_frames = malloc(img_in_md_buffer*sizeof(*md_frames));
    for(i=0; i<img_in_md_buffer; i++){
      md_frames[i].Y = calloc(current_frame.width*
			      current_frame.height,
			      sizeof(*md_frames[i].Y));
      if(current_frame.channels == 3){
	md_frames[i].Cr = calloc(current_frame.width*
				current_frame.height/2,
				sizeof(*md_frames[i].Cr));
	md_frames[i].Cb = calloc(current_frame.width*
				current_frame.height/2,
				sizeof(*md_frames[i].Cb));
      } else {
	md_frames[i].Cr = NULL;
	md_frames[i].Cb = NULL;
      }
      md_frames[i].channels = current_frame.channels;
      md_frames[i].width = current_frame.width;
      md_frames[i].height = current_frame.height;
    }
  }

  /* Frame struct used to store peak hold image generated from motion
     detection ring memory */
  if(md_peak_hold.Y == NULL){
    md_peak_hold.Y = (unsigned char*)malloc(current_frame.width *
					    current_frame.height *
					    sizeof(unsigned char));
    if(current_frame.channels == 3){
      md_peak_hold.Cr =  (unsigned char*)malloc(current_frame.width *
						current_frame.height *
						sizeof(unsigned char)/2);
      md_peak_hold.Cb =  (unsigned char*)malloc(current_frame.width *
						current_frame.height *
						sizeof(unsigned char)/2);
    } else {
      md_peak_hold.Cr = NULL;
      md_peak_hold.Cb = NULL;
    }
    md_peak_hold.channels = current_frame.channels;
    md_peak_hold.width = current_frame.width;
    md_peak_hold.height = current_frame.height;
  }

  /* Motion detection background arrays */
  md_trigger_threshold1 = (unsigned char*)calloc(current_frame.width*
						 current_frame.height,
						 sizeof(unsigned char));
  md_trigger_threshold2 = (unsigned char*)calloc(current_frame.width*
						 current_frame.height,
						 sizeof(unsigned char));
  
  /* Array of pixel luminances. These values are used in motion
     detection and in calculation of motion detection thresholds for
     each pixel */
  /* XXXXX luminances used for reference calculations. Filled upon
     request. Check naming! */
  md_luminance_calc = (unsigned char**)malloc(img_in_md_buffer*
					      sizeof(unsigned char*));
  for(i=0; i<img_in_md_buffer; i++){
    md_luminance_calc[i] = (unsigned char*)calloc(current_frame.width*
						  current_frame.height,
						  sizeof(unsigned char));
  }
  
  /* Point pointer to the median array currently in use */
  md_trigger_threshold = md_trigger_threshold1;

  /* Allocate and calculate lookup table for matrix -> vector indices */
  m2v = (int**)malloc(current_frame.height*sizeof(int*));
  for(i=0; i<current_frame.height; i++){
    m2v[i] = (int*)malloc(current_frame.width*sizeof(int));
    n = i * current_frame.width;
    for(j=0; j<current_frame.width; j++){
      m2v[i][j] = n + j;
    }
  }

  /* Allocate and calculate lookup tables for vector -> matrix
     indices. */
  v2m_x = (int*)malloc(current_frame.height*current_frame.width*sizeof(int));
  v2m_y = (int*)malloc(current_frame.height*current_frame.width*sizeof(int));

  for(i=0; i<current_frame.height*current_frame.width; i++){
    v2m_x[i] = i%current_frame.width;
    v2m_y[i] = (int)i/current_frame.width;
  }

  /* Allocate arrays for coordinates of pixels giving detection. */
  md_trigger_pixel_x = (int*)malloc(current_frame.height*current_frame.width*sizeof(int));
  md_trigger_pixel_y = (int*)malloc(current_frame.height*current_frame.width*sizeof(int));
  
  for(i=0; i<current_frame.height*current_frame.width; i++){
    md_trigger_pixel_x[i] = -1;
    md_trigger_pixel_y[i] = -1;
  }

}

void init_yuyv_lookups(void){
  int i, j, k;
  if(yuyv2r == NULL){

    /* Y-direction */
    yuyv2r = (unsigned char**)malloc(256*sizeof(unsigned char*));
    yuyv2g = (unsigned char***)malloc(256*sizeof(unsigned char*));
    yuyv2b = (unsigned char**)malloc(256*sizeof(unsigned char*));

    for(i=0; i<256; i++){
      yuyv2r[i] = (unsigned char*)malloc(256*sizeof(unsigned char));
      yuyv2g[i] = (unsigned char**)malloc(256*sizeof(unsigned char*));
      yuyv2b[i] = (unsigned char*)malloc(256*sizeof(unsigned char));

      for(j=0; j<256; j++){
	yuyv2g[i][j] = (unsigned char*)malloc(256*sizeof(unsigned char));
      }
    }

    /* Calculate lookup tables for each color, JPEG standard */
    for(i=0; i<256; i++){
      for(j=0; j<256; j++){
	/* R = limit(Y + 1.402 * (V - 128)) */
	yuyv2r[i][j] = limit(i + 1.402 * (j - 128));
	/* B = limit(Y + 1.772 * (U - 128)) */
	yuyv2b[i][j] = limit(i + 1.772 * (j - 128));

	for(k=0; k<256; k++){
	  /* G = limit(Y + 0.34414 * (U - 128) - (0.71414 * (V - 128))) */
	  yuyv2g[i][j][k] = limit(i + 0.34414 * (j - 128) - 0.71414 * (k - 128));
	}
      }
    }
  }
}

static void *md_process(void *ptr){

  int i = 0, j = 0, n = 0;
  unsigned int trigger_size = 0;
  int npix = current_frame.width*current_frame.height;
  int npix2 = 0.5*npix;

  //  int col_start, col_idx;
  //  int row_start, row_idx;
  int h_max = 0, w_max = 0;
  int trig_idx = -1;

  char *tmp = NULL;

  md_process_on = 1;

  /* Check if the current daily motion detection counter is up-to-date */
  /* Should this be here??? */
  if(md_trig_number == 0){
    char tmp3[FNAME_SIZE] = {'\0'};
    make_fname(tmp3,current_frame.timestamp,current_frame.usec,DAY_DIR);
    md_trig_number = find_md_folder_num(tmp3);
  }
  
  /* Update the timestamp */
  md_frames[md_i].timestamp = current_frame.timestamp;
  md_frames[md_i].usec = current_frame.usec;
  
  /* Copy current image buffer to ring memory */
  if(md_update_reference_buffers == 1){
    /* If md references should be updated, store data also to array
       used in calculations */
    for(i=0; i<npix; i++){
      md_frames[md_i].Y[i] = current_frame.Y[i];
      md_luminance_calc[md_i][i] = current_frame.Y[i];
    }
    frames_in_md_luminance_calc++;
  } else {
    for(i=0; i<npix; i++){
      md_frames[md_i].Y[i] = current_frame.Y[i];
    }
  }

  if(current_frame.channels == 3){
    for(i=0; i<npix2; i++){
      md_frames[md_i].Cr[i] = current_frame.Cr[i];
      md_frames[md_i].Cb[i] = current_frame.Cb[i];
    }
  }

  if(md_detection != -1){
    md_images_since_detection++;
  }

  if(frames_in_md_luminance_calc == img_in_md_buffer){
    frames_in_md_luminance_calc = 0;
    md_update_reference_buffers = 2;
    md_calc_references = 1;
    md_reference_buffer_ready = 1;
  }

  /* Detect rapid changes in illumination (motion detection) */
  if(md_references_ok == 1 && md_detection == -1 && md_ring_buffer_ready == 1){

    /*
      1. collect all indices, where the change has been above
      md_trigger_level

      2. calculate how many trigger pixels there is within the search
      box which has the upper-left corner at i

      3. profit

    */

    /* 1. */
    trigger_size = 0;
    for(i=0; i<npix; i++){
      if(current_frame.Y[i] >= md_trigger_threshold[i]){
	md_trigger_pixel_x[trigger_size] = v2m_x[i];
	md_trigger_pixel_y[trigger_size] = v2m_y[i];
	trigger_size++;
      }
    }
    n = trigger_size;

    /* If the whole image had too little valid points, there's no use
       go through the collected data. Inspect the data only if it is
       possible to get detection. */
    if(trigger_size >= md_trigger_size){
      //printf("n: %d, md: %d\n",n,md_trigger_size+1);
      for(i=0; i<n-md_trigger_size+1; i++){
	/* Maximum index values for the right and lower edges */
	w_max = md_trigger_pixel_x[i]+md_search_area[0];
	if(w_max > current_frame.width-1){ w_max = current_frame.width-1; }
	h_max = md_trigger_pixel_y[i]+md_search_area[1];
	if(h_max > current_frame.height-1){ w_max = current_frame.height-1; }

	//	printf("x: %d-%d, y: %d-%d\n", md_trigger_pixel_x[i], w_max,md_trigger_pixel_y[i], h_max);

	/* Check how many pixels there are within the search box which
	   has the top left corner located at the i:th pixel */
	trigger_size = 0;
	/* 2. */
	for(j=i+1; j<n; j++){
	  if(v2m_x[j] >= md_trigger_pixel_x[i] &&
	     v2m_x[j] < w_max &&
	     v2m_y[j] >= md_trigger_pixel_y[i] &&
	     v2m_y[j] < h_max){
	    /*
	      printf("i: %d, j: %d, v2m_x: %d, v2m_y: %d, w_max: %d, h_max: %d, md_x: %d, md_y: %d\n",
	      i,
	      j,
	      v2m_x[j],
	      v2m_y[j],
	      w_max,
	      h_max,
	      md_trigger_pixel_x[i],
	      md_trigger_pixel_y[i]);
	    */
	    trigger_size++;
	  }
	}
	/* If we got detection, stop going through the data. */
	if(trigger_size >= md_trigger_size){
	  trig_idx = i;
	  break;
	}
      }
    }

    if(trigger_size >= md_trigger_size){
      printf("ts %d, md_ts %d\n",trigger_size,md_trigger_size);
      if(md_detection == -1){
	md_detection = md_i;
	md_images_since_detection = 0;
      }
      tmp = make_time_string(current_frame.timestamp,current_frame.usec,TRIG);
      /* Get the date for the frame that triggers the saving. All the 
	 frames from this event are saved to same folder, no matter what 
	 date. */
      sscanf(tmp,"%d-%d-%d",&md_trig_year,&md_trig_month,&md_trig_day);

      if(save_log){
	if(md_log_fname == NULL){
	  md_log_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
	  memset(md_log_fname,'\0',FNAME_SIZE);
	}
	
	char tmp2[FNAME_SIZE] = {'\0'};
	
	make_fname(tmp2,current_frame.timestamp,current_frame.usec,MD_LOG_NAME);
	
	/* New file to be opened */
	if(md_log_fname[0] == '\0' || strcmp(tmp2, md_log_fname) != 0){
	  if(md_log_fid != NULL){
	    fclose(md_log_fid); md_log_fid = NULL;
	  }
	  
	  strcpy(md_log_fname,tmp2);
	  /* Add header if this is the first time we open the file
	   */
	  if(file_exists(md_log_fname)){
	    md_log_fid = fopen(md_log_fname,"a");
	  } else { // New file, add header and reset MD event counter
	    md_log_fid = fopen(md_log_fname,"a");
	    fprintf(md_log_fid,"# year, month, day, hour, minute, second, detection size, required detection size, required detection level, search area left edge, search area top edge, search area width, search area height, images saved, images saved before event, images saved after event, event number\n");
	    md_trig_number = 0;
	  }
	}
	
	if(tmp != NULL){ free(tmp); }
	
	tmp = make_time_string(current_frame.timestamp,current_frame.usec,MD_LOG_DATA);
	
	/* Now the file should be open until program exits or the 
	   day changes */
	
	fprintf(md_log_fid,"%s,%u,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
		tmp,
		trigger_size, /* Acquired trigger size */
		md_trigger_size, /* Required detection size */
		md_trigger_level, /* Required detection level */
		md_trigger_pixel_x[trig_idx], /* Left idx of the search area */
		md_trigger_pixel_y[trig_idx], /* Top idx of the search area */
		md_search_area[0], /* Width of the search area */
		md_search_area[1], /* Height of the search area */
		img_in_md_buffer,
		md_images_before,
		img_in_md_buffer-md_images_before-1,
		md_trig_number);
	fflush(md_log_fid);
	free(tmp);
      }

      i = 0;
      while(md_trigger_pixel_x[i] > -1){
	md_trigger_pixel_x[i] = -1;
	md_trigger_pixel_y[i] = -1;
      }
    }
  }
  
  // Increment index where to store image
  md_i++;
  // If we are using the last slot of the image buffer, start from zero
  if(md_i >= img_in_md_buffer){
    md_i = 0;
    md_ring_buffer_ready = 1;
  }
  
  md_process_on = 0;
  
  pthread_exit(NULL);

}

static void *calc_md_references(void *ptr){

  printf("calc_md_references\n");

  int n_pix = current_frame.width*current_frame.height;
  int n_imgs = img_in_md_buffer;

  int i, j;
  unsigned char med = 0;

  unsigned char *arr = (unsigned char*)malloc(n_imgs*sizeof(unsigned char));

  //  printf("Pixel medians\n");
  /* Calculate median value for each pixel */
  for(i=n_pix; i--;){
    if(md_mask[i] == 1){
      for(j=n_imgs; j--;){
	arr[j] = md_luminance_calc[j][i];
      }
      med = median_uint8(arr, n_imgs);
      if(med+md_trigger_level > 255){
	md_trigger_threshold_calc[i] = 255;
      } else {
	md_trigger_threshold_calc[i] = med + md_trigger_level;
      }

    } else {
      /* If the pixel is masked out, set median value to 255,
	 therefore making it impossible to getting trigger from this
	 location */
      md_trigger_threshold_calc[i] = 255;
    }
  }

  /* Now that the calculations are done, update the data to the new
     values */
  if(md_trigger_threshold == md_trigger_threshold1){
    md_trigger_threshold = md_trigger_threshold2;
    md_trigger_threshold_calc = md_trigger_threshold1;
  } else {
    md_trigger_threshold = md_trigger_threshold1;
    md_trigger_threshold_calc = md_trigger_threshold2;
  }

  /* Release local variables */
  free(arr);

  md_background_update_time = time(NULL);

  md_reference_calc_on = 0;
  md_calc_references = 0;
  md_update_reference_buffers = 0;
  md_references_ok = 1;

  printf("calc_md_references done\n");

  pthread_exit(NULL);

}

static void *md_save_images(void* ptr){

  char *fname = NULL;
  fname = (char*)malloc(FNAME_SIZE*sizeof(char));

  int i = 0, j = 0, k = 0;

  /* Save reference image */
  memset(fname,'\0',FNAME_SIZE);
  /*
  make_fname(fname,-1,-1,REFERENCE);
  writeImage(fname, 1, current_frame.width, current_frame.height, md_trigger_threshold, "spam!");
  */

  /* Save ring memory to disc */
  if(md_test == 0){
    printf("Saving ring memory..\n");
  }
  for(i=0; i<img_in_md_buffer; i++){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname,md_frames[i].timestamp, md_frames[i].usec, TRIG);
    if(md_test == 0){
      if(verbose == 1){
	printf("[%d] %s saved from ring memory\n",i,fname);
      }

      writeImage(fname, md_frames[i], "Finsprite motion detection at work");

      /* Update peak hold image */
      k=-1;
      for(j=0; j<md_frames[i].width*md_frames[i].height; j++){
	if(j%2 == 0){ k++; }
	if(md_peak_hold.Y[j] < md_frames[i].Y[j]){
	  md_peak_hold.Y[j] = md_frames[i].Y[j];
	  if(current_frame.channels == 3){
	    md_peak_hold.Cr[k] = md_frames[i].Cr[k];
	    md_peak_hold.Cb[k] = md_frames[i].Cb[k];
	  }
	}
      }
    }
  }

  if(md_test == 0){
    memset(fname,'\0',FNAME_SIZE);

    make_fname(fname,md_frames[md_i].timestamp, md_frames[md_i].usec, TRIG_PEAK);
    writeImage(fname, md_peak_hold,
	       "Finsprite motion detection at work");

    printf("%d images from ring memory saved.\n",img_in_md_buffer);
  }

  /* Clear the peak hold frame */
  k=-1;
  for(j=0; j<md_peak_hold.width*md_peak_hold.height; j++){
    if(j%2 == 0){ k++; }
    md_peak_hold.Y[j] = 0;
    if(md_peak_hold.channels == 3){
      md_peak_hold.Cr[k] = 0;
      md_peak_hold.Cb[k] = 0;
    }
  }

  md_trig_number++;

  md_images_since_detection = -1;
  md_detection = -1;

  /* Start to fill the motion detection ring memory from the beginning */
  md_i = 0;
  md_ring_buffer_ready = 0;

  /* Re-enable motion detection */
  md_saving_images = 0;

  free(fname); fname = NULL;

  pthread_exit(NULL);

}

static void *process_stack(void* ptr){

  int i = 0, j = 0;
  unsigned long min = ULONG_MAX;

  int n = current_frame.width*current_frame.height;
  int n2 = 0.5*n;

  /* Update stacks */

  if(use_max == 1 || use_dif == 1 || use_latest == 1){
    if((use_max == 1 || use_dif == 1) && use_latest == 1){
      if(peak_hold->Y == NULL){
	init_peak_hold();
      }
      if(latest_img->Y == NULL){
	init_latest_img();
      }
      for(i=0; i<n; i++){
	if(current_frame.Y[i] > peak_hold->Y[i]){
	  peak_hold->Y[i] = current_frame.Y[i];
	  latest_img->Y[i] = current_frame.Y[i];
	  if(current_frame.channels == 3 && i%2 == 0){
	    peak_hold->Cr[j] = current_frame.Cr[j];
	    peak_hold->Cb[j] = current_frame.Cb[j];
	    latest_img->Cr[j] = current_frame.Cr[j];
	    latest_img->Cb[j] = current_frame.Cb[j];
	    j++;
	  }
	}
      }
    } else if(use_max == 1 || use_dif == 1){
      if(peak_hold->Y == NULL){
	init_peak_hold();
      }
      for(i=0; i<n; i++){
	if(current_frame.Y[i] > peak_hold->Y[i]){
	  peak_hold->Y[i] = current_frame.Y[i];
	  if(current_frame.channels == 3 && i%2 == 0){
	    peak_hold->Cr[j] = current_frame.Cr[j];
	    peak_hold->Cb[j] = current_frame.Cb[j];
	    j++;
	  }
	}
      }
    } else {
      if(latest_img->Y == NULL){
	init_latest_img();
      }
      for(i=0; i<n; i++){
	if(current_frame.Y[i] > latest_img->Y[i]){
	  latest_img->Y[i] = current_frame.Y[i];
	  if(current_frame.channels == 3 && i%2 == 0){
	    latest_img->Cr[j] = current_frame.Cr[j];
	    latest_img->Cb[j] = current_frame.Cb[j];
	    j++;
	  }
	}
      }
    }
  }

  if(use_min == 1 || use_dif == 1){
    if(min_hold->Y == NULL){
      init_min_hold();
    }
    for(i=0; i<n; i++){
      if(current_frame.Y[i] < min_hold->Y[i]){
	min_hold->Y[i] = current_frame.Y[i];
	if(current_frame.channels == 3 && i%2 == 0){
	  min_hold->Cr[i] = current_frame.Cr[i];
	  min_hold->Cb[i] = current_frame.Cb[i];
	  j++;
	}
      }
    }
  }

  if(use_ave == 1){
    if(ave_stack->Y == NULL){
      init_ave_stack();
    }
    for(i=0; i<n; i++){
      ave_stack->Y[i] += current_frame.Y[i];
      if(ave_stack->Y[i] > ave_stack->max_val){
	ave_stack->max_val = ave_stack->Y[i];
      }
      if(ave_stack->Y[i] < min){
	min = ave_stack->Y[i];
      }
    }
    if(current_frame.channels == 3){
      for(i=0; i<n2; i++){
	ave_stack->Cr[i] += current_frame.Cr[i];
	ave_stack->Cb[i] += current_frame.Cb[i];
      }
    }
    ave_stack->num_imgs++;
    if(min > ave_stack->min_val){
      ave_stack->min_val = min;
    }
  }

  pthread_exit(NULL);

}

static int process_frame(const unsigned char *p, struct timeb tmb){

  int i = 0, j = 0, k = 0;

  current_frame.channels = number_of_channels;
  current_frame.width = fmt.fmt.pix.width;
  current_frame.height = fmt.fmt.pix.height;
  current_frame.timestamp = tmb.time;
  current_frame.usec = tmb.millitm*1000;

  if(current_frame.Y == NULL){
    current_frame.Y = malloc(current_frame.width*
			     current_frame.height*
			     sizeof(unsigned char*));
    if(current_frame.channels == 3){
      current_frame.Cr = malloc(current_frame.width*
				current_frame.height*
				sizeof(unsigned char)/2);
      current_frame.Cb = malloc(current_frame.width*
				current_frame.height*
				sizeof(unsigned char)/2);
    }
  }

  /* Assume YUYV format. RGB and GREY should be added? */
  switch(fmt.fmt.pix.pixelformat){
  case V4L2_PIX_FMT_GREY:
    for(i = 0, j = 0; i < fmt.fmt.pix.sizeimage; i+=4, j+=2){
      current_frame.Y[j+0] = p[j];
      current_frame.Y[j+1] = p[j+2];
    }
    break;
  case V4L2_PIX_FMT_UYVY:
    if(number_of_channels == 1){
      for(i = 0 ,j = 0; i < fmt.fmt.pix.sizeimage; i+=4, j+=2){
	current_frame.Y[j+0] = p[i+1];
	current_frame.Y[j+1] = p[i+3];
      }
    } else { // number_of_channels == 3
      for(i = 0, j = 0, k = 0; i < fmt.fmt.pix.sizeimage; i+=4, j+=2, k++){
	current_frame.Y[j+0] = p[i+1];
	current_frame.Y[j+1] = p[i+3];
	current_frame.Cb[k] = p[i];
	current_frame.Cr[k] = p[i+2];
      }
    }
    break;
  case V4L2_PIX_FMT_YUYV:
    if(number_of_channels == 1){
      for(i = 0 ,j = 0; i < fmt.fmt.pix.sizeimage; i+=4, j+=2){
	current_frame.Y[j+0] = p[i];
	current_frame.Y[j+1] = p[i+2];
      }
    } else { // number_of_channels == 3
      for(i = 0, j = 0, k = 0; i < fmt.fmt.pix.sizeimage; i+=4, j+=2, k++){
	current_frame.Y[j+0] = p[i];
	current_frame.Y[j+1] = p[i+2];
	current_frame.Cb[k] = p[i+1];
	current_frame.Cr[k] = p[i+3];
      }
    }
    
  default:
    break;
  }
  if(current_frame.Y[1000] > 0){ return 1; }
  else { return -1; }

}

static void *save_stacks(void *ptr){

  /* Save stack information to log */
  if(save_log){
    if(stack_log_fname == NULL){
      stack_log_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      memset(stack_log_fname,'\0',FNAME_SIZE);
    }

    char tmp[FNAME_SIZE] = {'\0'};
    char *tmp2 = NULL;

    make_fname(tmp,stack_timestamp,stack_usec,STACK_LOG_NAME);

    /* New file to be opened */
    if(stack_log_fname[0] == '\0' || strcmp(tmp, stack_log_fname) != 0){
      if(stack_log_fid != NULL){
	fclose(stack_log_fid); stack_log_fid = NULL;
      }

      strcpy(stack_log_fname,tmp);
      if(file_exists(stack_log_fname)){
	stack_log_fid = fopen(stack_log_fname,"a");
      } else {
	stack_log_fid = fopen(stack_log_fname,"a");
	fprintf(stack_log_fid,"# start year, start month, start day, start hour, start minute, start second, end year, end month, end day, end hour, end minute, end second, stack length, images in stack, min stack saved, max stack saved, ave stack saved, dif stack saved\n");
      }
    }

    tmp2 = make_time_string(stack_timestamp,stack_usec,STACK_LOG_DATA);
    fprintf(stack_log_fid,"%s,%d,%d,%d,%d,%d,%d\n",tmp2,stack_length,num_img_in_save_stack,use_min,use_max,use_ave,use_dif);
    fflush(stack_log_fid);

    free(tmp2);

    if(stack_log_fname != NULL){
      free(stack_log_fname);
      stack_log_fname = NULL;
    }
  }

  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);  

  if(use_ave == 1){
    save_ave_stack();
  }
  
  if(use_dif == 1){
    save_dif_stack();
  }

  if(use_max == 1){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname, peak_hold_save->timestamp, peak_hold_save->usec, PEAK_HOLD);
    printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
    //stretch(peak_hold->img_buffer, peak_hold->width*peak_hold->height, 0.01, 0.99);
    /*XXXX RGB conversion  */
    writeImage(fname, *peak_hold_save,
	       "Finsprite peak-hold stack");
  }
  
  if(use_max == 1 || use_dif == 1){
    free(peak_hold_save->Y); peak_hold_save->Y = NULL;
    if(current_frame.channels == 3){
      free(peak_hold_save->Cr); peak_hold_save->Cr = NULL;
      free(peak_hold_save->Cb); peak_hold_save->Cb = NULL;
    }
  }
  
  if(use_min == 1){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname, min_hold_save->timestamp, min_hold_save->usec, MIN_HOLD);
    printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
    //stretch(min_hold->img_buffer, min_hold->width*min_hold->height, 0.01, 0.99);
    /* Save image  */
    writeImage(fname, *min_hold_save, "Finsprite minimum stack");

  }
  
  if(use_min == 1 || use_dif == 1){
    free(min_hold_save->Y); min_hold_save->Y = NULL;
    if(current_frame.channels == 3){
      free(min_hold_save->Cr); min_hold_save->Cr = NULL;
      free(min_hold_save->Cb); min_hold_save->Cr = NULL;
    }
  }
  
  num_img_in_save_stack = 0;
  //  stack_timestamp = current_frame.timestamp;

  free(fname);

  saving_stacks = 0;

  pthread_exit(NULL);

}

static void *save_latest(void *ptr){

  saving_latest = 1;

  printf("Saving %s\n",latest_fname);
  writeImage(latest_fname, *latest_img_save, "Finsprite latest stack");
  
  free(latest_img_save->Y); latest_img_save->Y = NULL;
  if(current_frame.channels == 3){
    free(latest_img_save->Cr); latest_img_save->Cr = NULL;
    free(latest_img_save->Cb); latest_img_save->Cb = NULL;
  }

  saving_latest = 0;

  pthread_exit(NULL);

}

static void errno_exit(const char *s){
  fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
  exit(EXIT_FAILURE);
}

static int xioctl(int fd, int request, void *arg){
  int r;
  
  do {
    r = ioctl(fd, request, arg);
  } while(r == -1 && EINTR == errno);
  
  return r;
}

static char *make_time_string(time_t timestamp, long usec, int type){

  char *time_string = (char*)malloc(100*sizeof(char));
  char *tmp = NULL;

  if(type < MD_LOG_NAME){
    strftime(time_string,30,"%Y-%m-%d_%H%M%S",gmtime(&(timestamp)));
  }

  switch(type){
  case TRIG:
  case TRIG_PEAK:
    tmp = (char*)malloc(50*sizeof(char));
    sprintf(tmp,".%.3ld",usec/1000);
    strcat(time_string,tmp);
    free(tmp);
    break;
  case MD_LOG_NAME:
  case STACK_LOG_NAME:
    strftime(time_string,30,"%Y-%m-%d",gmtime(&(timestamp)));
    break;
  case MD_LOG_DATA:
  case STACK_LOG_DATA:
    tmp = (char*)malloc(30*sizeof(char));
    strftime(time_string,30,"%Y,%m,%d,%H,%M,%S",gmtime(&(timestamp)));
    sprintf(tmp,".%.3ld",usec/1000);
    strcat(time_string,tmp);
    if(type == STACK_LOG_DATA){
      memset(tmp,'\0',30);
      strftime(tmp,30,",%Y,%m,%d,%H,%M,%S",gmtime(&(stack_end_timestamp)));
      strcat(time_string,tmp);
      memset(tmp,'\0',30);
      sprintf(tmp,".%.3ld",stack_end_usec/1000);
      strcat(time_string,tmp);
    }
    free(tmp);
    break;
  }
  return time_string;
}

static void make_fname(char *fname, time_t timestamp, long usec, int type){
  char tmp[50] = {'\0'};
  mode_t dir_mode = S_IRWXU |S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

  if(out_dir != NULL){ strcat(fname,out_dir); strcat(fname,"/"); }

  if(type == TRIG || type == TRIG_PEAK || type == REFERENCE){
    sprintf(tmp,"%d",md_trig_year);
  } else {
    strftime(tmp,50,"%Y",gmtime(&(timestamp)));
  }
  strcat(fname,tmp);
  memset(tmp,'\0',50);

  if(mkdir(fname, dir_mode) < 0 && errno != EEXIST){
    printf("Unable to create %s\nSaving to current direcory: %d\n",fname,errno);
    // Empty path
    memset(fname,'\0',FNAME_SIZE);
  } else {

    if(type == TRIG || type == TRIG_PEAK || type == REFERENCE){
      sprintf(tmp,"/%.2d",md_trig_month);
    } else {
      strftime(tmp,50,"/%m",gmtime(&(timestamp)));
    }
    strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    // Save only images to daily subfolders
    if(type < MD_LOG_NAME){
      if(type == TRIG || type == TRIG_PEAK || type == REFERENCE){
	sprintf(tmp,"/%.2d",md_trig_day);
      } else {
	strftime(tmp,50,"/%d",gmtime(&(timestamp)));
      }
      strcat(fname,tmp); CLEAR(tmp);
      mkdir(fname, dir_mode);
    }

    /* Save motion detected images to event-folders, stacks to respective
       folders
    */
    switch(type){
    case MIN_HOLD:
      strcat(fname,"/min");
      break;
    case PEAK_HOLD:
      strcat(fname,"/max");
      break;
    case AVE:
      strcat(fname,"/ave");
      break;
    case DIF:
      strcat(fname,"/dif");
      break;
    case TRIG:
    case TRIG_PEAK:
    case REFERENCE:
      sprintf(tmp,"/MD%.4d",md_trig_number);
      strcat(fname,tmp);
      break;
    }

    mkdir(fname, dir_mode);
    memset(tmp,'\0',50);

    strcat(fname,"/");

  }

  /* If we want to have the daily image directory, skip filename creation */
  if(type != DAY_DIR){

    switch(type){
    case REFERENCE:
      strcat(fname,"reference.png");
      break;
    default:
      if(img_prefix != NULL){ strcat(fname,img_prefix); strcat(fname,"_"); }
      char *tmp2 = NULL;
      tmp2 = make_time_string(timestamp,usec,type);
      strcat(fname,tmp2); 
      free(tmp2);
      break;
    }
    
    switch(type){
    case AVE:
      sprintf(tmp,"_UTC_%d_s_ave.png", stack_length);
      break;
    case PEAK_HOLD:
      sprintf(tmp,"_UTC_%d_s_max.png", stack_length);
      break;
    case MIN_HOLD:
      sprintf(tmp,"_UTC_%d_s_min.png", stack_length);
      break;
    case DIF:
      sprintf(tmp,"_UTC_%d_s_dif.png", stack_length);
      break;
    case TRIG:
      sprintf(tmp,"_UTC_md.png");
      break;
    case TRIG_PEAK:
      sprintf(tmp,"_UTC_md_max.png");
      break;
    case MD_LOG_NAME:
      sprintf(tmp,"_md.log");
      break;
    case STACK_LOG_NAME:
      sprintf(tmp,"_stacks.log");
      break;
    case REFERENCE:
      break;
    default:
      sprintf(tmp,".%ld_UTC.png",usec);
      break;
    }
    strcat(fname,tmp);
  }

}

static void save_ave_stack(){

  int i, j;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);

  struct frame out_frame;
  out_frame.Y = NULL;
  out_frame.Cr = NULL;
  out_frame.Cb = NULL;

  out_frame.Y = calloc(ave_stack_save->width *
		       ave_stack_save->height,
		       sizeof(unsigned char));
  if(current_frame.channels == 3){
    out_frame.Cr = calloc(ave_stack_save->width *
		       ave_stack_save->height / 2,
		       sizeof(unsigned char));
    out_frame.Cb = calloc(ave_stack_save->width *
		       ave_stack_save->height / 2,
		       sizeof(unsigned char));
  }
  out_frame.channels = ave_stack_save->channels;
  out_frame.width = ave_stack_save->width;
  out_frame.height = ave_stack_save->height;
  out_frame.timestamp = ave_stack_save->timestamp;
  out_frame.usec = ave_stack_save->usec;

  j = -1;
  for(i=0; i<out_frame.width*out_frame.height; i++){
    out_frame.Y[i] = ave_stack_save->Y[i]/ave_stack_save->num_imgs;
    if(current_frame.channels == 3 && i%2 == 0){
      j++;
      out_frame.Cr[j] = ave_stack_save->Cr[j]/ave_stack_save->num_imgs;
      out_frame.Cb[j] = ave_stack_save->Cb[j]/ave_stack_save->num_imgs;
    }
    //out_frame[i] = 255*ave_stack_save->img_buffer[i]/ave_stack_save->max_val;
    //out_frame[i] = 255*(ave_stack_save->img_buffer[i]-ave_stack_save->min_val)/(ave_stack_save->max_val-ave_stack_save->min_val);
  }

  make_fname(fname, ave_stack_save->timestamp, ave_stack_save->usec, AVE);
  //stretch(out_frame, ave_stack_save->width*ave_stack_save->height, 0.02, 0.98);
  printf("Saving %s, %d frames used.\n",fname,ave_stack_save->num_imgs);
  writeImage(fname, out_frame, "Finsprite average stack");
  
  free(out_frame.Y);
  if(current_frame.channels == 3){
    free(out_frame.Cr);
    free(out_frame.Cb);
  }
  free(fname);

  free(ave_stack_save->Y); ave_stack_save->Y = NULL;
  if(current_frame.channels == 3){
    free(ave_stack_save->Cr); ave_stack_save->Cr = NULL;
    free(ave_stack_save->Cb); ave_stack_save->Cb = NULL;
  }
  ave_stack_save->min_val = ULONG_MAX;
  ave_stack_save->max_val = 0;
  ave_stack_save->num_imgs = 0;
}

static void save_dif_stack(){

  int i, j;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);

  //  unsigned char *dif = calloc(min_hold_save->npix,sizeof(float));

  struct frame out_frame;
  out_frame.Y = NULL;
  out_frame.Cr = NULL;
  out_frame.Cb = NULL;

  out_frame.Y = calloc(current_frame.width*current_frame.height,
		       sizeof(unsigned char));
  if(current_frame.channels == 3){
    out_frame.Cr = calloc(current_frame.width *
			  current_frame.height / 2,
			  sizeof(unsigned char));
    out_frame.Cb = calloc(current_frame.width *
			  current_frame.height / 2,
			  sizeof(unsigned char));
  }
  out_frame.channels = min_hold_save->channels;
  out_frame.width = min_hold_save->width;
  out_frame.height = min_hold_save->height;
  out_frame.timestamp = min_hold_save->timestamp;
  out_frame.usec = min_hold_save->usec;

  unsigned char dmin = UCHAR_MAX, dmax = 0;
  j=-1;
  for(i=0; i<out_frame.width*out_frame.height; i++){
    out_frame.Y[i] = peak_hold_save->Y[i] - min_hold_save->Y[i];
    if(current_frame.channels == 3){
      if(i%2 == 0){
	j++;
	out_frame.Cr[j] = peak_hold_save->Cr[j] - min_hold_save->Cr[j];
	out_frame.Cb[j] = peak_hold_save->Cb[j] - min_hold_save->Cb[j];
      }
    }
    if(out_frame.Y[i] < dmin){ dmin = out_frame.Y[i]; }
    if(out_frame.Y[i] > dmax){ dmax = out_frame.Y[i]; }
  }

  /* Scale luminance data to cover range 0-255 */
  for(i=0; i<out_frame.height*out_frame.width; i++){
    out_frame.Y[i] = 255*(out_frame.Y[i]-dmin)/dmax;
  }
  
  make_fname(fname, min_hold_save->timestamp, min_hold_save->usec, DIF);
  printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
  //stretch(out_frame, min_hold->width*min_hold->height, 0.02, 0.98);
  writeImage(fname, out_frame, "Finsprite difference stack");
  
  free(out_frame.Y); out_frame.Y = NULL;
  if(current_frame.channels == 3){
    free(out_frame.Cr); out_frame.Cr = NULL;
    free(out_frame.Cb); out_frame.Cb = NULL;
  }
  free(fname); fname = NULL;

}

static int read_frame(void){
  struct v4l2_buffer buf;

  time_t frametime;

  struct timeb tmb;

  int status = 0, i;
  void *status2 = NULL;

  CLEAR(buf);

  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
    switch(errno) {
    case EAGAIN:
      printf("EAGAIN\n");
      return 0;
      
    case EIO:
      /* Could ignore EIO, see spec. */
      printf("EIO\n");
      /* fall through */
      
    default:
      printf("default\n");
      errno_exit("VIDIOC_DQBUF");
    }
  }
    
  assert(buf.index < n_mmap_buffers);
  
  frametime = buf.timestamp.tv_sec;

  if(frametime > 0){
    ftime(&tmb);
    frametime = time(NULL);
    if(process_frame(mmap_buffers[buf.index].start, tmb) > 0){

      if(use_trig == 1 && md_luminance_calc == NULL){
	init_md_arrays();
      }
      /* Skip N frames from the beginning of imaging period */
      if(delay_start > 0){
	delay_start--;
      } else {
	num_img_in_stack++;

	//      pthread_t stack_thread, trig_thread;
	//pthread_attr_t thread_attr;

	int stack_thread_used = 0, trig_thread_used = 0;

	/* Check statuses of motion detection stuff, if it is enabled */
	if(use_trig == 1){
	  /*XXXXXXXXX don't let this stop updating the ring buffers! */
	  /* If ring memory is ready to be saved, create a new thread for it */
	  if(md_saving_images == 0 &&  md_images_since_detection > md_images_before){

	    /* Disable motion detection until the ring memory is being saved */
	    md_saving_images = 1;

	    status = pthread_create(&trig_save_thread, &thread_attr_detached, md_save_images, NULL);
	    if(status){
	      printf("Thread \"md_save_images()\" failed with code %d\n",status);
	    }
	  }

	  /* Check if it is time to update the motion detection
	     reference data */
	  if(md_update_reference_buffers == 0 &&
	     //md_reference_buffer_ready == 0 && 
	     md_threshold_update_interval <= 
	     difftime(frametime, md_background_update_time)){
	    md_update_reference_buffers = 1;
	    //md_reference_buffer_ready = 0;
	    //md_calc_references = 1;
	  }

	  /* If motion detection reference data buffer is full and
	     ready for use
	     matrices */
	  if(md_reference_buffer_ready == 1 &&
	     md_reference_calc_on == 0 &&
	     md_calc_references == 1){

	    if(md_trigger_threshold == md_trigger_threshold1){
	      md_trigger_threshold = md_trigger_threshold2;
	      md_trigger_threshold_calc = md_trigger_threshold1;
	    } else {
	      md_trigger_threshold = md_trigger_threshold1;
	      md_trigger_threshold_calc = md_trigger_threshold2;
	    }

	    /* Copy current threshold data to matrices used while new
	       references are being calculated */
	    if(md_background_update_time > 0){
	      for(i=0; i<current_frame.width*current_frame.height; i++){
		md_trigger_threshold[i] = md_trigger_threshold_calc[i];
	      }
	    }

	    md_reference_calc_on = 1;

	    status = pthread_create(&md_references_thread, &thread_attr_detached, calc_md_references, NULL);
	  }

	}
	
	if(use_trig == 1 && md_saving_images == 0){
	  trig_thread_used = 1;
	  status = pthread_create(&md_trig_thread, &thread_attr_join, md_process, NULL);
	  //status = pthread_create(&md_trig_thread, NULL, md_process, NULL);
	  if(status){
	    printf("Thread \"md_process()\" failed with code %d\n",status);
	    trig_thread_used = 0;
	  }
	}

	if(use_max == 1 || use_min == 1 || use_ave == 1 || use_dif == 1 || use_latest == 1){
	  
	  stack_thread_used = 1;
	  status = pthread_create(&stack_thread, &thread_attr_join, process_stack, NULL);
	  //status = pthread_create(&stack_thread, NULL, process_stack, NULL);
	  if(status){
	    printf("Thread \"process_stack()\" failed with code %d\n",status);
	    stack_thread_used = 0;
	  }
	}

	/* Join threads */
	
	/* Stacking should be faster than motion detection, so attach
	   it first */

	if(stack_thread_used){
	  status = pthread_join(stack_thread, &status2);
	  if(status) {
	    printf("ERROR; return code from pthread_join() for stack_thread is %d\n", status);
	  }
	  stack_thread_used = 0;
	}

	if(trig_thread_used){
	  status = pthread_join(md_trig_thread, &status2);
	  if(status) {
	    printf("ERROR; return code from pthread_join() for stack_thread is %d\n", status);
	  }
	  trig_thread_used = 0;
	}
      
	if((use_ave == 1 || use_min == 1 || use_max == 1 || use_dif == 1) &&
	   stack_length < difftime(current_frame.timestamp,stack_timestamp)){

	  stack_end_timestamp = current_frame.timestamp;
	  stack_end_usec = current_frame.usec;

	  num_img_in_save_stack = num_img_in_stack;
	  num_img_in_stack = 0;

	  /* Swap stack pointers and create thread for stack saving */
	  if(min_hold == &min_hold1){
	    min_hold = &min_hold2;
	    min_hold_save = &min_hold1;
	  } else {
	    min_hold = &min_hold1;
	    min_hold_save = &min_hold2;
	  }
	  if(peak_hold == &peak_hold1){
	    peak_hold = &peak_hold2;
	    peak_hold_save = &peak_hold1;
	  } else {
	    peak_hold = &peak_hold1;
	    peak_hold_save = &peak_hold2;
	  }
	  if(ave_stack == &ave_stack1){
	    ave_stack = &ave_stack2;
	    ave_stack_save = &ave_stack1;
	  } else {
	    ave_stack = &ave_stack1;
	    ave_stack_save = &ave_stack2;
	  }

	  saving_stacks = 1;

	  status = pthread_create(&stack_save_thread, &thread_attr_detached, save_stacks, NULL);
	  if(status){
	    printf("Thread \"save_stacks()\" failed with code %d\n",status);
	  }

	}

	if(use_latest == 1 && 
	   latest_length < difftime(current_frame.timestamp,latest_img_timestamp)){

	  if(latest_img == &latest_img1){
	    latest_img = &latest_img2;
	    latest_img_save = &latest_img1;
	  } else {
	    latest_img = &latest_img1;
	    latest_img_save = &latest_img2;
	  }

	  status = pthread_create(&latest_save_thread, &thread_attr_detached, save_latest, NULL);
	  if(status){
	    printf("Thread \"save_latest()\" failed with code %d\n",status);
	  }

	} // latest.png saving

      }
    } // if process_frame got good data

  } /* if valid frame time was received */

  if(capture_length != 0){
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }

  return 1;
}

static void mainloop(void){
  time_t timenow;

  start_time = time(NULL);
  timenow = time(NULL);

  /**
  char *time_string = malloc(100*sizeof(char));
  strftime(time_string,30,"%Y-%m-%d_%H%M%S",gmtime(&(timenow)));
  printf("%s\n",time_string);
  free(time_string);
  */

  while(difftime(timenow,start_time) < capture_length) {

    for(;;) {
      fd_set fds;
      struct timeval tv;
      int r;
      
      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      
      /* Timeout. */
      tv.tv_sec = 5;
      tv.tv_usec = 0;
      
      r = select(fd + 1, &fds, NULL, NULL, &tv);
      
      if(r == -1) {
	if(EINTR == errno)
	  continue;
	
	errno_exit("select");
      }
      
      if(r == 0) {
	fprintf(stderr, "select timeout\n");
	exit(EXIT_FAILURE);
      }
      
      if(read_frame())
	break;
      
      /* EAGAIN - continue select loop. */
    }
    timenow = time(NULL);
  }

}

static void stop_capturing(void){
  enum v4l2_buf_type type;
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMOFF, &type) == -1)
    errno_exit("VIDIOC_STREAMOFF");

}

static void start_capturing(void){
  unsigned int i;
  enum v4l2_buf_type type;
  
  for(i = 0; i < n_mmap_buffers; ++i) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = i;
    
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMON, &type) == -1){
    errno_exit("VIDIOC_STREAMON");
  }
}

static void uninit_device(void){
  unsigned int i;
  
  for(i = 0; i < n_mmap_buffers; ++i)
    if(munmap(mmap_buffers[i].start, mmap_buffers[i].length) == -1)
      errno_exit("munmap");
  
  free(mmap_buffers);
}

static void init_mmap(int n){
  struct v4l2_requestbuffers req;
  
  CLEAR(req);
  
  req.count = 2;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s does not support memory mapping\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_REQBUFS");
    }
  }
  
  if(req.count < 2) {
    fprintf(stderr, "Insufficient buffer memory on %s\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  mmap_buffers = calloc(req.count, sizeof(*mmap_buffers));
  
  if(!mmap_buffers) {
    fprintf(stderr, "Out of memory\n");
    exit(EXIT_FAILURE);
  }
  
  for(n_mmap_buffers = 0; n_mmap_buffers < req.count; ++n_mmap_buffers) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n_mmap_buffers;
    
    if(xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
      errno_exit("VIDIOC_QUERYBUF");
    
    mmap_buffers[n_mmap_buffers].length = buf.length;
    mmap_buffers[n_mmap_buffers].start =
      mmap(NULL /* start anywhere */,
	   buf.length,
	   PROT_READ | PROT_WRITE /* required */,
	   MAP_SHARED /* recommended */,
	   fd, buf.m.offset);
    
    if(MAP_FAILED == mmap_buffers[n_mmap_buffers].start)
      errno_exit("mmap");
  }
}

static void init_device(int img_in_mmap_buffer){
  struct v4l2_capability cap;
  struct v4l2_cropcap cropcap;
  struct v4l2_crop crop;
  unsigned int min;
  
  if(xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s is no V4L2 device\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_QUERYCAP");
    }
  }
  
  if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    fprintf(stderr, "%s is no video capture device\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
    fprintf(stderr, "%s does not support streaming i/o\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  /* Select video input, video standard and tune here. */
  
  CLEAR(cropcap);

  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0) {
    crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c = cropcap.defrect; /* reset to default */
    
    if(xioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
      switch(errno) {
      case EINVAL:
	/* Cropping not supported. */
	break;
      default:
	/* Errors ignored. */
	break;
      }
    }
  } else {
    /* Errors ignored. */
  }
    
  CLEAR(fmt);
  
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = 720;
  fmt.fmt.pix.height = 576;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

  if(xioctl(fd, VIDIOC_S_FMT, &fmt) == -1)
    errno_exit("VIDIOC_S_FMT");
  
  /* Note VIDIOC_S_FMT may change width and height. */
  
  /* Buggy driver paranoia. */
  min = fmt.fmt.pix.width * 2;
  if(fmt.fmt.pix.bytesperline < min)
    fmt.fmt.pix.bytesperline = min;
  min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
  if(fmt.fmt.pix.sizeimage < min)
    fmt.fmt.pix.sizeimage = min;
  
  init_mmap(img_in_mmap_buffer);

}

static void close_device(void){
  if(close(fd) == -1)
    errno_exit("close");
  
  fd = -1;
}

static void open_device(void){
  struct stat st; 
  
  if(stat(dev_name, &st) == -1) {
    fprintf(stderr, "Cannot identify '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  if(!S_ISCHR(st.st_mode)) {
    fprintf(stderr, "%s is no device\n", dev_name);
    exit(EXIT_FAILURE);
  }
  
  fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
  
  if(fd == -1) {
    fprintf(stderr, "Cannot open '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
}

static void usage(FILE *fp, int argc, char **argv){
  fprintf(fp,
	  "Usage: %s [options]\n\n"
	  "Options:\n"
	  "-c | --capture-device <name> Video device name [/dev/video0]\n"
	  "-h | --help  Print this message\n"
	  "-n | --buffer-size <num>  Number of images in ring buffer [25]\n"
	  "                          Enables motion detection\n"
	  "-l | --capture-length <num>  Length of imaging period in\n"
	  "                             seconds [default: unlimited]\n"
	  "-s | --stack-length <num>  Length of stacked period in seconds [60]\n"
	  "-L | --latest-length <num>  Length of imaging period for \"latest.png\" in\n"
	  "                            seconds [default: 60]\n"
	  "-m | --min  Enable minimum stacking [default: off]\n"
	  "-M | --max  Enable maximum stacking [default: off]\n"
	  "-a | --ave  Enable average stacking [default: off]\n"
	  "-d | --dif  Enable difference stacking [default: off]\n"
	  "-t | --trig  Enable motion detection with default ring buffer [default: off]\n"
	  "-T | --trigger-level-and-size <num,num>  Minimum trigger level and size\n"
	  "                                         required for motion detection [10,20]\n"
	  "-S | --search_area  Define the width and height of the area which should\n"
	  "                                         contain the triggering event [10,10]\n"
	  "-e | --latest  Save \"latest.png\", an independent peak-hold stack\n"
	  "-U | --update-interval <num>  Interval between motion detection background\n"
	  "                              field calculation in seconds [default: 60]\n"
	  "-k | --mask <file>  Mask filename used to exclude areas from motion detect\n"
	  "-G | --trigger-test  Test how sensitive triggering is. No images are saved\n"
	  "-B | --images-before <num>  Number of images to be saved that were recorded\n"
	  "                            before the triggering event [default: \n"
	  "                            (number of images - 1)/2]\n"
	  "-D | --delay-start <num>  Delay start of imaging for <num> frames. Useful\n"
	  "                          eg. for cameras requiring time to adjust for\n"
	  "                          brightness or white balance [default: 10]\n"
	  "-p | --prefix <prefix>  Image filename prefix [default: none]\n"
	  "-o | --out-dir <dir>  Image output directory [default: current directory]\n"
	  "                      The images are saved to <dir>/yyyy/mm/dd/\n"
	  "-O | --latest-fname <file>  Filename with full path for saving\n"
	  "                            \"latest.png\" [default: latest.png]\n"
	  "-N | --no-log  Turn of logging\n"
	  "-C | --number_of_channels <num>  Number of color channels (1 or 3) [default: 1]\n"
	  "-f | --fps <num>  FPS requested from the video device [default: auto]\n"
	  "\n",
	  argv[0]);
}

int main(int argc, char **argv){
  dev_name = "/dev/video0";

  int i;

  for(;;) {
    int index;
    int c;
    
    c = getopt_long(argc, argv,
		    short_options, long_options,
		    &index);
    
    if(c == -1)
      break;
    
    switch(c) {
    case 0: /* getopt_long() flag */
      break;
    case 'c':
      dev_name = optarg;
      break;      
    case 'h':
      usage(stdout, argc, argv);
      exit(EXIT_SUCCESS);
    case 'n':
      sscanf(optarg,"%d",&img_in_md_buffer);
      use_trig = 1;
      break;
    case 's':  /* length of stacked period in secs */
      sscanf(optarg,"%d",&stack_length);
      break;
    case 'L': /* Length of "latest.png" stack */
      sscanf(optarg,"%d",&latest_length);
      break;
    case 'm': /* Enable minimum stack */
      use_min = 1;
      break;
    case 'M': /* Enable maximum stack */
      use_max = 1;
      break;
    case 'a': /* Enable average stack */
      use_ave = 1;
      break;
    case 'S':
      sscanf(optarg,"%d,%d",&md_search_area[0],&md_search_area[1]);
    case 't': /* Enable triggering (motion detection)*/
      use_trig = 1;
      break;
    case 'e': /* Enable saving of "latest.png" */
      use_latest = 1;
      break;
    case 'T': /* Minimum trigger/motion detect threshold*/
      sscanf(optarg,"%hhu,%ud",&md_trigger_level,&md_trigger_size);
      break;
    case 'U': /* Interval between background calculations */
      sscanf(optarg,"%d",&md_threshold_update_interval);
      break;
    case 'k': /* Mask that can be used to exclude pixels from motion detection */
      md_mask_fname = optarg;
      break;
    case 'G':
      md_test = 1;
      use_trig = 1;
      break;
    case 'B':
      sscanf(optarg,"%d",&md_images_before);
      break;
    case 'D':
      sscanf(optarg,"%ud",&delay_start);
      break;
    case 'd': /* Enable difference stack */
      use_dif = 1;
      break;
    case 'l': /* Length of imaging period */
      sscanf(optarg,"%ld",&capture_length);
      break;
    case 'p': /* Image prefix */
      img_prefix = optarg;
      break;
    case 'o': /* Output directory prefix */
      out_dir = optarg;
      break;
    case 'O': /* Filename for "latest.png" */
      latest_fname = optarg;
      use_latest = 1;
      break;
    case 'N': /* Disable logging */
      save_log = 0;
      break;
    case 'C':
      sscanf(optarg,"%d",&number_of_channels);
      break;
    case 'f':
      sscanf(optarg,"%d",&video_input_fps);
      break;
    case 'V': /* Verbose */
      verbose = 1;
      break;
    default:
      usage(stderr, argc, argv);
      exit(EXIT_FAILURE);
    }
  }

  if(use_min == 0 && use_max == 0 && use_ave == 0 && use_trig == 0 && 
     use_dif == 0 && use_latest == 0){
    printf("\nNothing to do, terminating.\n");
    printf("Specify at least one imaging mode!\n\n");
    usage(stderr, argc, argv);
    exit(EXIT_FAILURE);
  }

  if(md_test){
    use_min = 0;
    use_max = 0;
    use_ave = 0;
    use_dif = 0;
    use_latest = 0;
  }

  if(number_of_channels != 1 && number_of_channels != 3){
    number_of_channels = 1;
  }
  if(number_of_channels == 3){
    //init_yuyv_lookups();
  }

  if(use_latest == 1 && latest_fname == NULL){
    latest_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
    strcpy(latest_fname,latest_fname_default);
  }

  if(md_images_before == -1 || md_images_before > img_in_md_buffer){
    md_images_before = (img_in_md_buffer-1)/2;
  }

  ave_stack1.Y = NULL;
  ave_stack1.Cr = NULL;
  ave_stack1.Cb = NULL;
  ave_stack1.min_val = ULONG_MAX;
  ave_stack1.max_val = 0;

  ave_stack2.Y = NULL;
  ave_stack2.Cr = NULL;
  ave_stack2.Cb = NULL;
  ave_stack2.min_val = ULONG_MAX;
  ave_stack2.max_val = 0;

  ave_stack = &ave_stack1;
  ave_stack_save = NULL;

  peak_hold1.Y = NULL;
  peak_hold2.Y = NULL;
  peak_hold1.Cr = NULL;
  peak_hold2.Cr = NULL;
  peak_hold1.Cb = NULL;
  peak_hold2.Cb = NULL;
  peak_hold = &peak_hold1;
  peak_hold_save = NULL;

  min_hold1.Y = NULL;
  min_hold2.Y = NULL;
  min_hold1.Cr = NULL;
  min_hold2.Cr = NULL;
  min_hold1.Cb = NULL;
  min_hold2.Cb = NULL;
  min_hold = &min_hold1;
  min_hold_save = NULL;

  latest_img1.Y = NULL;
  latest_img2.Y = NULL;
  latest_img1.Cr = NULL;
  latest_img2.Cr = NULL;
  latest_img1.Cb = NULL;
  latest_img2.Cb = NULL;
  latest_img = &latest_img1;
  latest_img_save = NULL;

  md_peak_hold.Y = NULL;
  md_peak_hold.Cr = NULL;
  md_peak_hold.Cb = NULL;

  pthread_attr_init(&thread_attr_join);
  pthread_attr_setdetachstate(&thread_attr_join, PTHREAD_CREATE_JOINABLE);

  pthread_attr_init(&thread_attr_detached);
  pthread_attr_setdetachstate(&thread_attr_detached, PTHREAD_CREATE_DETACHED);
  
  open_device();

  init_device(img_in_mmap_buffer);

  start_capturing();

  mainloop();

  /* Wait until all threads have finished */
  while(md_process_on == 1 || md_reference_calc_on == 1 || md_saving_images == 1 || saving_stacks == 1 || saving_latest == 1){
  }
  
  pthread_attr_destroy(&thread_attr_join);
  pthread_attr_destroy(&thread_attr_detached);

  if(capture_length != 0){
    stop_capturing();

    uninit_device();

    close_device();
  }

  if(use_trig == 1){
    for(i=0; i < img_in_md_buffer; i++){
      if(md_frames[i].Y != NULL){
	free(md_frames[i].Y);
	md_frames[i].Y = NULL;
	if(md_frames[i].Cr != NULL){
	  free(md_frames[i].Cr); md_frames[i].Cr = NULL;
	}
	if(md_frames[i].Cb != NULL){
	  free(md_frames[i].Cb); md_frames[i].Cb = NULL;
	}
      }
    }
  }
  if(md_frames != NULL){ free(md_frames); }

  if(peak_hold1.Y != NULL){ free(peak_hold1.Y); }
  if(peak_hold1.Cr != NULL){ free(peak_hold1.Cr); }
  if(peak_hold1.Cb != NULL){ free(peak_hold1.Cb); }
  if(peak_hold2.Y != NULL){ free(peak_hold2.Y); }
  if(peak_hold2.Cr != NULL){ free(peak_hold2.Cr); }
  if(peak_hold2.Cb != NULL){ free(peak_hold2.Cb); }
  if(min_hold1.Y != NULL){ free(min_hold1.Y); }
  if(min_hold1.Cr != NULL){ free(min_hold1.Cr); }
  if(min_hold1.Cb != NULL){ free(min_hold1.Cb); }
  if(min_hold2.Y != NULL){ free(min_hold2.Y); }
  if(min_hold2.Cr != NULL){ free(min_hold2.Cr); }
  if(min_hold2.Cb != NULL){ free(min_hold2.Cb); }
  if(ave_stack1.Y != NULL){ free(ave_stack1.Y); }
  if(ave_stack1.Cr != NULL){ free(ave_stack1.Cr); }
  if(ave_stack1.Cb != NULL){ free(ave_stack1.Cb); }
  if(ave_stack2.Y != NULL){ free(ave_stack2.Y); }
  if(ave_stack2.Cr != NULL){ free(ave_stack2.Cr); }
  if(ave_stack2.Cr != NULL){ free(ave_stack2.Cb); }
  if(latest_img1.Y != NULL){ free(latest_img1.Y); }
  if(latest_img1.Cr != NULL){ free(latest_img1.Cr); }
  if(latest_img1.Cb != NULL){ free(latest_img1.Cb); }
  if(latest_img2.Y != NULL){ free(latest_img2.Y); }
  if(latest_img2.Cr != NULL){ free(latest_img2.Cr); }
  if(latest_img2.Cb != NULL){ free(latest_img2.Cb); }

  if(md_trigger_threshold1 != NULL){ free(md_trigger_threshold1); }
  if(md_trigger_threshold2 != NULL){ free(md_trigger_threshold2); }
  if(md_luminance_calc != NULL){
    for(i=0; i<img_in_md_buffer; i++){ free(md_luminance_calc[i]); }
    free(md_luminance_calc);
  }

  if(m2v != NULL){
    for(i=0; i<current_frame.height; i++){ free(m2v[i]); }
    free(m2v);
  }

  if(v2m_x != NULL){ free(v2m_x); }
  if(v2m_y != NULL){ free(v2m_y); }
  if(md_trigger_pixel_x != NULL){ free(md_trigger_pixel_x); }
  if(md_trigger_pixel_y != NULL){ free(md_trigger_pixel_y); }

  if(current_frame.Y != NULL){ free(current_frame.Y); }
  if(current_frame.Cr != NULL){ free(current_frame.Cr); }
  if(current_frame.Cb != NULL){ free(current_frame.Cb); }

  if(md_log_fid != NULL){ fclose(md_log_fid); }
  if(stack_log_fid != NULL){ fclose(stack_log_fid); }

  if(md_mask != NULL){ free(md_mask); }

  exit(EXIT_SUCCESS);

  return 0;
}
